import asyncio
from collections import namedtuple
from contextlib import asynccontextmanager
from datetime import date, datetime, time, timedelta
from functools import wraps
from pathlib import Path
import random
import sys
import site
import traceback
from typing import Any, Coroutine, Iterable, Union

import click
from loguru import logger
from typer import Typer
from typer.core import TyperCommand

from . import var, __url__, __name__

Flagged = namedtuple("Flagged", ("noflag", "flag"))


def get_path_frame(e, path):
    try:
        tb = traceback.extract_tb(e.__traceback__)
        for frame in reversed(tb):
            if Path(path) in Path(frame.filename).parents:
                return frame
        else:
            return None
    except AttributeError:
        return None


def get_last_frame(e):
    try:
        tb = traceback.extract_tb(e.__traceback__)
        for frame in reversed(tb):
            return frame
    except AttributeError:
        return None


def get_cls_fullpath(c):
    module = c.__module__
    if module == "builtins":
        return c.__qualname__
    return module + "." + c.__qualname__


def format_exception(e, regular=True):
    if regular:
        prompt = "\n请在 Github 或交流群反馈下方错误详情以帮助开发者修复该问题:\n"
    else:
        prompt = ""
    proj_path = Path(__file__).parent.absolute()
    proj_frame = get_path_frame(e, proj_path)
    if proj_frame:
        proj_frame_path = Path(proj_frame.filename).relative_to(proj_path)
        prompt += f"\n  P {proj_frame_path}, L {proj_frame.lineno}, F {proj_frame.name}:"
        prompt += f"\n    {proj_frame.line.strip()}"
    last_frame = get_last_frame(e)
    if last_frame:
        last_frame_path = last_frame.filename
        for p in site.getsitepackages():
            if Path(p) in Path(last_frame.filename).parents:
                last_frame_path = "<SP>/" + str(Path(last_frame.filename).relative_to(p))
                break
        prompt += f"\n  S {last_frame_path}, L {last_frame.lineno}, F {last_frame.name}:"
        prompt += f"\n    {last_frame.line.strip()}"
    prompt += f"\n    E {get_cls_fullpath(type(e))}: {e}\n"
    return prompt


def show_exception(e, regular=True):
    if (regular and 1 < var.debug < 2) or (not regular and var.debug < 2):
        var.console.rule()
        print(format_exception(e, regular=regular), flush=True, file=sys.stderr)
        var.console.rule()
    else:
        logger.opt(exception=e).debug("错误详情:")


class AsyncTyper(Typer):
    """Typer 的异步版本, 所有命令函数都将以异步形式调用."""

    def async_command(self, *args, **kwargs):
        def decorator(async_func):
            @wraps(async_func)
            def sync_func(*_args, **_kwargs):
                try:
                    asyncio.run(async_func(*_args, **_kwargs))
                except KeyboardInterrupt:
                    print("\r", end="", flush=True)
                    logger.info(f"所有客户端已停止, 欢迎您再次使用 {__name__.capitalize()}.")
                except Exception as e:
                    print("\r", end="", flush=True)
                    logger.critical(f"发生关键错误, {__name__.capitalize()} 将退出.")
                    show_exception(e, regular=False)
                    sys.exit(1)
                else:
                    logger.info(f"所有任务已完成, 欢迎您再次使用 {__name__.capitalize()}.")

            self.command(*args, **kwargs)(sync_func)
            return async_func

        return decorator


class FlagValueCommand(TyperCommand):
    """允许在命令行参数中使用"="."""

    def parse_args(self, ctx, args):
        long = {}
        short = {}
        defined = set()
        for o in self.params:
            if isinstance(o, click.Option):
                if isinstance(o.default, Flagged):
                    for pre in o.opts:
                        if pre.startswith("--"):
                            long[pre] = o
                        elif pre.startswith("-"):
                            short[pre] = o

        for i, a in enumerate(args):
            a = a.split("=")
            if a[0] in long:
                defined.add(long[a[0]])
                if len(a) == 1:
                    args[i] = f"{a[0]}={long[a[0]].default.flag}"
            elif a[0] in short:
                defined.add(short[a[0]])
                if len(args) == i + 1 or args[i + 1].startswith("-"):
                    args.insert(i + 1, str(short[a[0]].default.flag))

        for u in set(long.values()) - defined:
            for p, o in long.items():
                if o == u:
                    break
            args.append(f"{p}={u.default.noflag}")

        return super().parse_args(ctx, args)


class AsyncTaskPool:
    """一个用于批量等待异步任务的管理器, 支持在等待时添加任务."""

    def __init__(self):
        self.waiter = asyncio.Condition()
        self.tasks = []

    def add(self, coro: Coroutine):
        async def wrapper():
            task = asyncio.ensure_future(coro)
            await asyncio.wait([task])
            async with self.waiter:
                self.waiter.notify()
                return await task

        t = asyncio.create_task(wrapper())
        t.set_name(coro.__name__)
        self.tasks.append(t)
        return t

    async def as_completed(self):
        for t in self.tasks:
            if t.done():
                yield t
                self.tasks.remove(t)
        while self.tasks:
            async with self.waiter:
                await self.waiter.wait()
                for t in self.tasks[:]:
                    if t.done():
                        yield t
                        self.tasks.remove(t)

    async def wait(self):
        results = []
        async for t in self.as_completed():
            results.append(t.result())
        return results


class AsyncCountPool(dict):
    """
    一个异步安全的 ID 分配器.
    参数:
        base: ID 起始数
    """

    def __init__(self, *args, base=1000, **kw):
        super().__init__(*args, **kw)
        self.lock = asyncio.Lock()
        self.next = base + 1

    async def append(self, value):
        """输入一个值, 该值将被存储并分配一个 ID."""
        async with self.lock:
            key = self.next
            self[key] = value
            self.next += 1
            return key


def to_iterable(var: Union[Iterable, Any]):
    """
    将任何变量变为可迭代变量.
    说明:
        None 将变为空数组.
        非可迭代变量将变为仅有该元素的长度为 1 的数组.
        可迭代变量将保持不变.
    """
    if var is None:
        return ()
    if isinstance(var, str) or not isinstance(var, Iterable):
        return (var,)
    else:
        return var


def remove_prefix(text: str, prefix: str):
    """从字符串头部去除前缀."""
    return text[text.startswith(prefix) and len(prefix) :]


def truncate_str(text: str, length: int):
    """将字符串截断到特定长度, 并增加"..."后缀."""
    return f"{text[:length + 3]}..." if len(text) > length else text


def time_in_range(start, end, x):
    """判定时间在特定范围内."""
    if start <= end:
        return start <= x <= end
    else:
        return start <= x or x <= end


def batch(iterable, n=1):
    """将数组分成 N 份."""
    l = len(iterable)
    for ndx in range(0, l, n):
        yield iterable[ndx : min(ndx + n, l)]


def flatten(l):
    """将二维数组变为一维数组."""
    return [item for sublist in l for item in sublist]


def async_partial(f, *args1, **kw1):
    """Partial 函数的异步形式."""

    async def func(*args2, **kw2):
        return await f(*args1, *args2, **kw1, **kw2)

    return func


async def idle():
    """异步无限等待函数."""
    await asyncio.Event().wait()


def random_time(start_time: time = None, end_time: time = None):
    """在特定的开始和结束时间之间生成时间, 如果开始时间晚于结束时间, 视为过夜."""
    start_datetime = datetime.combine(date.today(), start_time or time(0, 0))
    end_datetime = datetime.combine(date.today(), end_time or time(23, 59, 59))
    if end_datetime < start_datetime:
        end_datetime += timedelta(days=1)
    time_diff_seconds = (end_datetime - start_datetime).total_seconds()
    random_seconds = random.randint(0, int(time_diff_seconds))
    random_time = (start_datetime + timedelta(seconds=random_seconds)).time()
    return random_time


def next_random_datetime(start_time: time = None, end_time: time = None, interval_days=1):
    """在特定的开始和结束时间之间生成时间, 并设定最小间隔天数."""
    min_date = (datetime.now() + timedelta(days=interval_days)).date()
    min_datetime = datetime.combine(min_date, time(0, 0))
    target_time = random_time(start_time, end_time)
    offset_date = 0
    while True:
        t = datetime.combine(datetime.now() + timedelta(days=offset_date), target_time)
        if t >= min_datetime:
            break
        else:
            offset_date += 1
    return t


def humanbytes(B: float):
    """将字节数转换为人类可读形式."""
    """Return the given bytes as a human friendly KB, MB, GB, or TB string."""
    B = float(B)
    KB = float(1024)
    MB = float(KB**2)  # 1,048,576
    GB = float(KB**3)  # 1,073,741,824
    TB = float(KB**4)  # 1,099,511,627,776

    if B < KB:
        return "{0} {1}".format(B, "Bytes" if 0 == B > 1 else "Byte")
    elif KB <= B < MB:
        return "{0:.2f} KB".format(B / KB)
    elif MB <= B < GB:
        return "{0:.2f} MB".format(B / MB)
    elif GB <= B < TB:
        return "{0:.2f} GB".format(B / GB)
    elif TB <= B:
        return "{0:.2f} TB".format(B / TB)


@asynccontextmanager
async def no_waiting(lock: asyncio.Lock):
    try:
        await asyncio.wait_for(lock.acquire(), 0)
    except asyncio.TimeoutError:
        acquired = False
    else:
        acquired = True
    try:
        yield
    finally:
        if acquired:
            lock.release()


def distribute_numbers(min_value, max_value, num_elements=1, min_distance=0, max_distance=None, base=[]):
    if max_value < min_value:
        raise ValueError("invalid value range.")

    if max_distance and max_distance < min_distance:
        raise ValueError("invalid distance range.")

    numbers = sorted(base)
    results = []

    for _ in range(num_elements):
        allowed_range = []
        for i in range(-1, len(numbers)):
            if i == -1:
                min_allowed_value = min_value
            else:
                min_allowed_value = max(numbers[i] + min_distance, min_value)
            if i == len(numbers) - 1:
                max_allowed_value = max_value
            else:
                max_allowed_value = min(numbers[i + 1] - min_distance, max_value)
            if min_allowed_value < max_allowed_value:
                allowed_range.append((min_allowed_value, max_allowed_value))
        if not allowed_range:
            break
        estimated_num_elements = [
            min(int((r[1] - r[0]) // min_distance), num_elements) for r in allowed_range
        ]
        r = random.choices(allowed_range, k=1, weights=estimated_num_elements)[0]
        d = r[1] - r[0]
        min_v = r[0] + min_distance if r[0] == min_value else r[0]
        max_v = r[1]
        if max_distance and d > max_distance:
            value = random.uniform(min_v, r[0] + max_distance - min_distance)
        else:
            value = random.uniform(min_v, max_v)
        numbers = sorted(numbers + [value])
        results.append(value)
    return sorted(results)



import os
import pty
import select
import fcntl
import struct
import subprocess
import sys
import termios
import threading
import time
import signal

import typer
from loguru import logger
from flask import Flask, render_template, request, redirect, url_for, jsonify, abort
from flask_socketio import SocketIO
from flask_login import LoginManager, login_user, login_required, current_user

from . import __version__

cli = typer.Typer()
app = Flask(__name__, static_folder="templates/assets")
app.config["SECRET_KEY"] = os.urandom(24)
socketio = SocketIO(app, cors_allowed_origins="*")
login_manager = LoginManager()
login_manager.init_app(app)
login_manager.login_view = "login"

app.config["args"] = None
app.config["fd"] = None
app.config["pid"] = None
app.config["hist"] = ""
app.config["faillog"] = []

lock = threading.Lock()

version = f"V{__version__}"


class DummyUser:
    def is_authenticated(self):
        return True

    def is_active(self):
        return True

    def is_anonymous(self):
        return False

    def get_id(self):
        return 0


@login_manager.user_loader
def load_user(_):
    return DummyUser()


@app.route("/")
def index():
    return redirect(url_for("console"))


@app.route("/console")
@login_required
def console():
    return render_template("console.html", version=version)


@app.route("/login", methods=["GET"])
def login():
    return render_template("login.html", version=version)


@app.route("/login", methods=["POST"])
def login_submit():
    password = request.form.get("password", "")
    webpass = os.environ.get("EK_WEBPASS", "")
    if not webpass:
        emsg = "Web console password not set."
    elif sum(t > time.time() - 3600 for t in app.config["faillog"][-5:]) == 5:
        emsg = "Too many login trials in one hour."
    else:
        if password == webpass:
            login_user(DummyUser())
            return redirect(request.args.get("next") or url_for("index"))
        else:
            emsg = "Wrong password."
            app.config["faillog"].append(time.time())
    return render_template("login.html", emsg=emsg, version=version)


@app.route("/healthz")
def healthz():
    return "200 OK"


@app.route("/heartbeat")
def heartbeat():
    webpass = os.environ.get("EK_WEBPASS", "")
    password = request.args.get("pass", None)
    if (not password) or (not webpass):
        return abort(403)
    if password == webpass:
        if app.config["pid"] is None:
            start({"rows": 32, "cols": 106}, auth=False)
            return jsonify({"status": "restarted", "pid": app.config["pid"]}), 201
        else:
            return jsonify({"status": "running", "pid": app.config["pid"]}), 200
    else:
        return abort(403)


@app.errorhandler(404)
def page_not_found(e):
    return render_template("404.html", version=version), 404


@socketio.on("pty-input", namespace="/pty")
def pty_input(data):
    if not current_user.is_authenticated():
        return
    if app.config["fd"]:
        os.write(app.config["fd"], data["input"].encode())


def set_size(fd, row, col, xpix=0, ypix=0):
    logger.debug(f"Resizing pty to: {row} {col}.")
    size = struct.pack("HHHH", row, col, xpix, ypix)
    fcntl.ioctl(fd, termios.TIOCSWINSZ, size)


@socketio.on("resize", namespace="/pty")
def resize(data):
    if not current_user.is_authenticated():
        return
    if app.config["fd"]:
        set_size(app.config["fd"], data["rows"], data["cols"])


def read_and_forward_pty_output():
    max_read_bytes = 1024 * 20
    while True:
        socketio.sleep(0.01)
        if app.config["fd"]:
            (data, _, _) = select.select([app.config["fd"]], [], [], 0)
            if data:
                output = os.read(app.config["fd"], max_read_bytes).decode(errors="ignore")
                app.config["hist"] += output
                socketio.emit("pty-output", {"output": output}, namespace="/pty")


@socketio.on("connect", namespace="/pty")
def connected():
    logger.debug(f"Console connected.")


@socketio.on("disconnect", namespace="/pty")
def connected():
    logger.debug(f"Console disconnected.")


@socketio.on("embykeeper_start", namespace="/pty")
def start(data, auth=True):
    if auth and (not current_user.is_authenticated()):
        return
    with lock:
        if app.config["fd"]:
            set_size(app.config["fd"], data["rows"], data["cols"])
            socketio.emit("pty-output", {"output": app.config["hist"]}, namespace="/pty")
        else:
            (pid, fd) = pty.fork()
            if pid == 0:
                while True:
                    try:
                        p = subprocess.run(["embykeeper", *app.config["args"]])
                        if p.returncode:
                            raise RuntimeError()
                    except (KeyboardInterrupt, RuntimeError):
                        print()
                        while True:
                            try:
                                input("Embykeeper 已退出, 请按 Enter 以重新开始 ...")
                                print()
                                break
                            except KeyboardInterrupt:
                                print()
            else:
                app.config["fd"] = fd
                app.config["pid"] = pid
                logger.debug(f"Embykeeper started at: {pid}.")
                set_size(app.config["fd"], data["rows"], data["cols"])
                socketio.start_background_task(target=read_and_forward_pty_output)


@socketio.on("embykeeper_kill", namespace="/pty")
def stop():
    if not current_user.is_authenticated():
        return
    with lock:
        if app.config["pid"] is not None:
            os.kill(app.config["pid"], signal.SIGINT)
            for _ in range(50):
                try:
                    os.kill(app.config["pid"], 0)
                except OSError:
                    break
                else:
                    time.sleep(0.1)
            else:
                os.kill(app.config["pid"], signal.SIGKILL)
            app.config["fd"] = None
            app.config["pid"] = None
            app.config["hist"] = ""
            logger.debug(f"Embykeeper stopped.")


@cli.command(context_settings={"ignore_unknown_options": True, "allow_extra_args": True})
def run(
    ctx: typer.Context,
    port: int = typer.Option(1818, envvar="PORT", show_envvar=False),
    host: str = "0.0.0.0",
    debug: bool = False,
):
    app.config["args"] = ctx.args
    logger.info(f"Embykeeper webserver started at {host}:{port}.")
    socketio.run(app, port=port, host=host, debug=debug)


if __name__ == "__main__":
    cli()



from datetime import datetime
from ..lock import pornemby_nohp
from .base import Messager


class PornembyMessager(Messager):
    name = "Pornemby"
    chat_name = "Pornemby"
    default_messages = ["pornemby-common-wl@latest.yaml * 100"]
    additional_auth = ["pornemby_pack"]

    async def prepare_send(self, message):
        nohp_date = pornemby_nohp.get(self.me.id, None)
        if nohp_date and nohp_date >= datetime.today().date():
            self.log.info(f"取消发送: 血量已耗尽.")
        else:
            return message



import asyncio
import time
from pathlib import Path
from typing import Iterable, Union

import aiofiles
import aiohttp
from aiohttp_socks import ProxyConnector, ProxyType
from cachetools import TTLCache
from loguru import logger

from .utils import humanbytes, no_waiting, show_exception, to_iterable

logger = logger.bind(scheme="datamanager")

cdn_urls = [
    "https://raw.githubusercontent.com/embykeeper/embykeeper-data/main",
    "https://raw.gitmirror.com/embykeeper/embykeeper-data/main",
    "https://cdn.jsdelivr.net/gh/embykeeper/embykeeper-data",
]

versions = TTLCache(maxsize=128, ttl=600)
lock = asyncio.Lock()


async def refresh_version(connector):
    async with no_waiting(lock):
        for data_url in cdn_urls:
            url = f"{data_url}/version"
            async with aiohttp.ClientSession(connector=connector) as session:
                try:
                    async with session.get(url) as resp:
                        if resp.status == 200:
                            result = await resp.text()
                            for l in result.splitlines():
                                if l:
                                    a, b = l.split("=")
                                    versions[a.strip()] = b.strip()
                            break
                        else:
                            logger.warning(f"资源文件版本信息获取失败 ({resp.status})")
                            return False
                except aiohttp.ClientConnectorError as e:
                    continue
                except Exception as e:
                    logger.warning(f"资源文件版本信息获取失败 ({e})")
                    show_exception(e)
                    return False
        else:
            logger.warning(f"资源文件版本信息获取失败.")
            return False


async def get_datas(basedir: Path, names: Union[Iterable[str], str], proxy: dict = None, caller: str = None):
    """
    获取额外数据.
    参数:
        basedir: 文件存储默认位置
        names: 要下载的路径列表
        proxy: 代理配置
        caller: 请求下载的模块名, 用于消息提示
    """
    basedir.mkdir(parents=True, exist_ok=True)

    existing = {}
    not_existing = []
    for name in to_iterable(names):
        if (basedir / name).is_file():
            logger.debug(f'检测到请求的本地文件: "{name}".')
            existing[name] = basedir / name
        else:
            not_existing.append(name)

    if not_existing:
        logger.info(f"{caller or '该功能'} 正在下载或更新资源文件: {', '.join(not_existing)}")

    for name in to_iterable(names):
        version_matching = False
        while True:
            if (basedir / name).is_file():
                yield basedir / name
            else:
                try:
                    for data_url in cdn_urls:
                        url = f"{data_url}/data/{name}"
                        if proxy:
                            connector = ProxyConnector(
                                proxy_type=ProxyType[proxy["scheme"].upper()],
                                host=proxy["hostname"],
                                port=proxy["port"],
                            )
                        else:
                            connector = aiohttp.TCPConnector()
                        async with aiohttp.ClientSession(connector=connector) as session:
                            try:
                                async with session.get(url) as resp:
                                    if resp.status == 200:
                                        file_size = int(resp.headers.get("Content-Length", 0))
                                        logger.info(f"开始下载: {name} ({humanbytes(file_size)})")
                                        async with aiofiles.open(basedir / name, mode="wb+") as f:
                                            timer = time.time()
                                            length = 0
                                            async for chunk in resp.content.iter_chunked(512):
                                                if time.time() - timer > 3:
                                                    timer = time.time()
                                                    logger.info(
                                                        f"正在下载: {name} ({humanbytes(length)} / {humanbytes(file_size)})"
                                                    )
                                                await f.write(chunk)
                                                length += len(chunk)
                                        logger.info(f"下载完成: {name} ({humanbytes(file_size)})")
                                        yield basedir / name
                                        break
                                    elif resp.status in (403, 404) and not version_matching:
                                        await refresh_version(connector=connector)
                                        if name in versions:
                                            logger.debug(f'解析版本 "{name}" -> "{versions[name]}"')
                                            name = versions[name]
                                            version_matching = True
                                            break
                                        else:
                                            logger.warning(f"下载失败: {name} ({resp.status})")
                                            yield None
                                            break
                                    else:
                                        logger.warning(f"下载失败: {name} ({resp.status})")
                                        yield None
                                        break
                            except aiohttp.ClientConnectorError as e:
                                (basedir / name).unlink(missing_ok=True)
                                continue
                            except Exception as e:
                                (basedir / name).unlink(missing_ok=True)
                                logger.warning(f"下载失败: {name} ({e})")
                                show_exception(e)
                                yield None
                                break
                    else:
                        logger.warning(f"下载失败: {name}.")
                        yield None
                        continue
                except KeyboardInterrupt:
                    (basedir / name).unlink(missing_ok=True)
                    raise
            if not version_matching:
                break


async def get_data(basedir: Path, name: str, proxy: dict = None, caller: str = None):
    async for data in get_datas(basedir, name, proxy, caller):
        return data



import asyncio
from urllib.parse import parse_qs, urlencode, urlparse

from aiohttp import ClientSession
from aiohttp_socks import ProxyConnector
from pyrogram.raw.functions.messages import RequestWebView
from pyrogram.raw.functions.users import GetFullUser
from faker import Faker

from ...utils import remove_prefix
from ..link import Link
from .base import BaseBotCheckin

__ignore__ = True


class NebulaCheckin(BaseBotCheckin):
    name = "Nebula"
    bot_username = "Nebula_Account_bot"

    def __init__(self, *args, **kw):
        super().__init__(*args, **kw)
        self.failed = False
        self.timeout *= 3

    async def fail(self):
        self.failed = True
        self.finished.set()

    async def start(self):
        try:
            try:
                await asyncio.wait_for(self._checkin(), self.timeout)
            except asyncio.TimeoutError:
                pass
        except OSError as e:
            self.log.info(f'发生错误: "{e}".')
            return False
        if not self.finished.is_set():
            self.log.warning("无法在时限内完成签到.")
            return False
        else:
            return not self.failed

    async def _checkin(self):
        bot = await self.client.get_users(self.bot_username)
        self.log.info(f"开始执行签到: [green]{bot.name}[/] [gray50](@{bot.username})[/].")
        bot_peer = await self.client.resolve_peer(self.bot_username)
        user_full = await self.client.invoke(GetFullUser(id=bot_peer))
        url = user_full.full_user.bot_info.menu_button.url
        url_auth = (
            await self.client.invoke(RequestWebView(peer=bot_peer, bot=bot_peer, platform="ios", url=url))
        ).url
        scheme = urlparse(url_auth)
        data = remove_prefix(scheme.fragment, "tgWebAppData=")
        url_base = scheme._replace(path="/api/proxy/userCheckIn", query=f"data={data}", fragment="").geturl()
        scheme = urlparse(url_base)
        query = parse_qs(scheme.query, keep_blank_values=True)
        query = {k: v for k, v in query.items() if not k.startswith("tgWebApp")}
        token, proxy, useragent = await Link(self.client).captcha()
        if (not token) or (not proxy):
            self.log.warning("签到失败: 无法获得验证码.")
            return await self.fail()
        if not useragent:
            useragent = Faker().safari()
        query["token"] = token
        url_checkin = scheme._replace(query=urlencode(query, True)).geturl()
        connector = ProxyConnector.from_url(proxy)
        async with ClientSession(connector=connector) as session:
            async with session.get(url_checkin, headers={"User-Agent": useragent}) as resp:
                results = await resp.json()
            message = results["message"]
            if any(s in message for s in ("未找到用户", "权限错误")):
                self.log.info("签到失败: 账户错误.")
                await self.fail()
            if "失败" in message:
                self.log.info("签到失败.")
                await self.fail()
            if "重复" in message:
                self.log.info("今日已经签到过了.")
                self.finished.set()
            elif "成功" in message:
                self.log.info(f"[yellow]签到成功[/]: + {results['get_credit']} 分 -> {results['credit']} 分.")
                self.finished.set()
            else:
                self.log.warning(f"接收到异常返回信息: {message}")



from .base import BotCheckin

__ignore__ = True


class BlueseaCheckin(BotCheckin):
    name = "Bluesea"
    bot_username = "blueseamusic_bot"
    bot_checkin_cmd = ["/cancel", "/checkin"]
    bot_captcha_len = 4
    bot_text_ignore = ["会话已取消", "没有活跃的会话"]



import asyncio
from pathlib import Path
import random
import re
from abc import ABC, abstractmethod
from contextlib import asynccontextmanager
from enum import Flag, auto
import string
import time
from typing import Iterable, List, Union

from ddddocr import DdddOcr
from appdirs import user_data_dir
from loguru import logger
from PIL import Image
from pyrogram import filters
from pyrogram.errors import UsernameNotOccupied, FloodWait
from pyrogram.handlers import EditedMessageHandler, MessageHandler
from pyrogram.types import InlineKeyboardMarkup, Message, ReplyKeyboardMarkup
from onnxruntime.capi.onnxruntime_pybind11_state import InvalidProtobuf
from thefuzz import fuzz

from ...data import get_datas
from ...utils import show_exception, to_iterable, AsyncCountPool
from ..lock import ocrs, ocrs_lock
from ..tele import Client
from ..link import Link

__ignore__ = True

default_keywords = {
    "account_fail": (
        "拉黑",
        "黑名单",
        "冻结",
        "未找到用户",
        "无资格",
        "退出群",
        "退群",
        "加群",
        "加入群聊",
        "请先关注",
        "请先加入",
        "注册",
        "不存在",
    ),
    "too_many_tries_fail": ("已尝试", "过多"),
    "checked": ("只能", "已经", "下次", "过了", "签过", "明日再来", "上次签到"),
    "fail": ("失败", "错误", "超时"),
    "success": ("成功", "通过", "完成", "获得"),
}


class MessageType(Flag):
    TEXT = auto()
    CAPTION = auto()
    CAPTCHA = auto()
    ANSWER = auto()


class BaseBotCheckin(ABC):
    """基础签到类."""

    name = None

    def __init__(
        self, client: Client, retries=4, timeout=120, nofail=True, basedir=None, proxy=None, config: dict = {}
    ):
        """
        基础签到类.
        参数:
            client: Pyrogram 客户端
            retries: 最大重试次数
            timeout: 签到超时时间
            nofail: 启用错误处理外壳, 当错误时报错但不退出
            basedir: 文件存储默认位置
            proxy: 代理配置
            config: 当前签到器的特定配置
        """
        self.client = client
        self.retries = retries
        self.timeout = timeout
        self.nofail = nofail
        self.basedir = basedir or user_data_dir(__name__)
        self.proxy = proxy
        self.config = config
        self.finished = asyncio.Event()  # 签到完成事件
        self.log = logger.bind(scheme="telechecker", name=self.name, username=client.me.name)  # 日志组件

    async def _start(self):
        """签到器的入口函数的错误处理外壳."""
        try:
            return await self.start()
        except asyncio.CancelledError:
            raise
        except Exception as e:
            if self.nofail:
                self.log.warning(f"初始化错误, 签到器将停止.")
                show_exception(e, regular=False)
                return False
            else:
                raise

    @abstractmethod
    async def start(self):
        pass


class BotCheckin(BaseBotCheckin):
    """签到类, 用于回复模式签到."""

    group_pool = AsyncCountPool(base=2000)
    ocr = None

    name: str = None  # 签到器的名称
    bot_username: Union[int, str] = None  # Bot 的 UserID 或 用户名 (不带 @ 或 https://t.me/)
    bot_checkin_cmd: Union[str, List[str]] = ["/checkin"]  # Bot 依次执行的签到命令
    bot_send_interval: int = 1  # 签到命令间等待的秒数
    bot_checkin_caption_pat: str = None  # 当 Bot 返回图片时, 仅当符合该 regex 才识别为验证码, 置空不限制
    bot_text_ignore: Union[str, List[str]] = []  # 当含有列表中的关键词, 即忽略该消息, 置空不限制
    bot_captcha_len: Union[int, Iterable[int]] = []  # 验证码的可能范围, 例如 [1, 2, 3], 置空不限制
    bot_success_pat: str = r"(\d+)[^\d]*(\d+)"  # 当接收到成功消息后, 从消息中提取数字的模式
    bot_retry_wait: int = 2  # 失败时等待的秒数
    bot_use_history: int = None  # 首先尝试识别历史记录中最后一个验证码图片, 最多识别 N 条, 置空禁用
    bot_allow_from_scratch: bool = False  # 允许从未聊天情况下启动
    bot_success_keywords: Union[str, List[str]] = []  # 成功时检测的关键词 (暂不支持regex), 置空使用内置关键词表
    bot_checked_keywords: Union[str, List[str]] = []  # 今日已签到时检测的关键词, 置空使用内置关键词表
    bot_account_fail_keywords: Union[str, List[str]] = []  # 账户错误将退出时检测的关键词 (暂不支持regex), 置空使用内置关键词表
    bot_too_many_tries_fail_keywords: Union[str, List[str]] = []  # 账户错误将退出时检测的关键词 (暂不支持regex), 置空使用内置关键词表
    bot_fail_keywords: Union[str, List[str]] = []  # 签到错误将重试时检测的关键词 (暂不支持regex), 置空使用内置关键词表
    chat_name: str = None  # 在群聊中向机器人签到
    additional_auth: List[str] = []  # 额外认证要求

    def __init__(self, *args, **kw):
        super().__init__(*args, **kw)
        self._is_archived = False  # 签到前处于收纳状态
        self._retries = 0  # 当前重试次数
        self._waiting = {}  # 当前等待的消息

    def get_filter(self):
        """设定要签到的目标."""
        filter = filters.user(self.bot_username)
        if self.chat_name:
            filter = filter & filters.chat(self.chat_name)
        else:
            filter = filter & filters.private
        return filter

    def get_handlers(self):
        """设定要监控的更新的类型."""
        return [
            MessageHandler(self._message_handler, self.get_filter()),
            EditedMessageHandler(self._message_handler, self.get_filter()),
        ]

    @asynccontextmanager
    async def listener(self):
        """执行监控上下文."""
        group = await self.group_pool.append(self)
        handlers = self.get_handlers()
        for h in handlers:
            self.client.add_handler(h, group=group)
        yield
        for h in handlers:
            try:
                self.client.remove_handler(h, group=group)
            except ValueError:
                pass

    async def get_ocr(self, ocr: str = None):
        """加载特定标签的 OCR 模型, 默认加载 ddddocr 默认模型."""
        while True:
            async with ocrs_lock:
                if ocr in ocrs:
                    return ocrs[ocr]
                if not ocr:
                    model = DdddOcr(beta=True, show_ad=False)
                else:
                    data = []
                    files = (f"{ocr}.onnx", f"{ocr}.json")
                    async for p in get_datas(self.basedir, files, proxy=self.proxy, caller=self.name):
                        if p is None:
                            self.log.warning(f"初始化错误: 无法下载所需文件.")
                            return None
                        else:
                            data.append(p)
                    try:
                        model = DdddOcr(
                            show_ad=False, import_onnx_path=str(data[0]), charsets_path=str(data[1])
                        )
                    except InvalidProtobuf:
                        self.log.warning(f"文件下载不完全, 正在重试下载.")
                        Path(str(data[0])).unlink()
                        Path(str(data[1])).unlink()
                        continue
                ocrs[ocr] = model
                return model

    async def start(self):
        """签到器的入口函数."""
        ident = self.chat_name or self.bot_username
        while True:
            try:
                chat = await self.client.get_chat(ident)
            except UsernameNotOccupied:
                self.log.warning(f'初始化错误: 会话 "{ident}" 不存在.')
                return False
            except KeyError as e:
                self.log.info(f"初始化错误: 无法访问, 您可能已被封禁: {e}.")
                show_exception(e)
                return False
            except FloodWait as e:
                self.log.info(f"初始化信息: Telegram 要求等待 {e.value} 秒.")
                if e.value < 360:
                    await asyncio.sleep(e.value)
                else:
                    self.log.info(f"初始化信息: Telegram 要求等待 {e.value} 秒, 您可能操作过于频繁, 签到器将停止.")
                    return False
            else:
                break
        async for d in self.client.get_dialogs(folder_id=1):
            if d.chat.id == chat.id:
                self._is_archived = True
                break
        else:
            async for d in self.client.get_dialogs(folder_id=0):
                if d.chat.id == chat.id:
                    break
            else:
                if not self.bot_allow_from_scratch:
                    self.log.info(f'跳过签到: 从未与 "{ident}" 交流.')
                    return None

        if self.additional_auth:
            for a in self.additional_auth:
                if not await Link(self.client).auth(a):
                    self.log.info(f"初始化错误: 权限校验不通过, 需要: {a}.")
                    return False

        bot = await self.client.get_users(self.bot_username)
        msg = f"开始执行签到: [green]{bot.name}[/] [gray50](@{bot.username})[/]"
        if chat.title:
            msg += f" @ [green]{chat.title}[/] [gray50](@{chat.username})[/]"
        self.log.info(msg + ".")

        if not self.chat_name:
            self.log.debug(f"[gray50]禁用提醒 {self.timeout} 秒: {bot.username}[/]")
            await self.client.mute_chat(ident, time.time() + self.timeout + 10)
        if await self.init():
            try:
                async with self.listener():
                    cancelled = False
                    try:
                        if self.bot_use_history is None:
                            await self.send_checkin()
                        elif not await self.walk_history(self.bot_use_history):
                            await self.send_checkin()
                        await asyncio.wait_for(self.finished.wait(), self.timeout)
                    except asyncio.CancelledError:
                        cancelled = True
                        raise
                    finally:
                        if not cancelled:
                            if not await self.cleanup():
                                self.log.debug(f"[gray50]执行清理失败: {ident}[/]")
                            if self._is_archived:
                                self.log.debug(f"[gray50]将会话重新归档: {ident}[/]")
                                try:
                                    if await chat.archive():
                                        self.log.debug(f"[gray50]重新归档成功: {ident}[/]")
                                except asyncio.TimeoutError:
                                    self.log.debug(f"[gray50]归档失败: {ident}[/]")
                            if not self.chat_name:
                                self.log.debug(f"[gray50]将会话设为已读: {ident}[/]")
                                try:
                                    if await self.client.read_chat_history(ident):
                                        self.log.debug(f"[gray50]设为已读成功: {ident}[/]")
                                except asyncio.TimeoutError:
                                    self.log.debug(f"[gray50]设为已读失败: {ident}[/]")
            except OSError as e:
                self.log.warning(f'初始化错误: "{e}", 签到器将停止.')
                show_exception(e)
                return False
            except asyncio.TimeoutError:
                pass
        else:
            self.log.warning(f"初始化错误.")
            return False

        if not self.finished.is_set():
            self.log.warning("无法在时限内完成签到.")
            return False
        else:
            return self._retries <= self.retries

    async def init(self):
        """可重写的初始化函数, 在读取聊天后运行, 在执行签到前运行, 返回 False 将视为初始化错误."""
        return True

    async def cleanup(self):
        """可重写的签到结束后的清理函数, 执行签到成功或失败后运行, 返回 False 将视为清理错误."""
        return True

    async def walk_history(self, limit=0):
        """处理 limit 条历史消息, 并检测是否有验证码."""
        try:
            async for m in self.client.get_chat_history(self.chat_name or self.bot_username, limit=limit):
                if MessageType.CAPTCHA in self.message_type(m):
                    await self.on_photo(m)
                    return True
            return False
        except Exception as e:
            self.log.warning("读取历史消息失败, 将不再读取历史消息.")
            show_exception(e)
            return False

    async def send(self, cmd):
        """向机器人发送命令."""
        if self.chat_name:
            bot = await self.client.get_users(self.bot_username)
            await self.client.send_message(self.chat_name, f"{cmd}@{bot.username}")
        else:
            await self.client.send_message(self.bot_username, cmd)

    async def send_checkin(self, retry=False):
        """发送签到命令, 或依次发送签到命令序列."""
        cmds = to_iterable(self.bot_checkin_cmd)
        for i, cmd in enumerate(cmds):
            if retry and not i:
                await asyncio.sleep(self.bot_retry_wait)
            if i < len(cmds):
                await asyncio.sleep(self.bot_send_interval)
            await self.send(cmd)

    async def _message_handler(self, client: Client, message: Message):
        """消息处理入口函数的错误处理外壳."""
        try:
            await self.message_handler(client, message)
        except OSError as e:
            self.log.info(f'发生错误: "{e}", 正在重试.')
            await self.retry()
            message.continue_propagation()
        except asyncio.CancelledError:
            raise
        except Exception as e:
            if not self.nofail:
                await self.fail()
                raise
            else:
                self.log.warning(f"发生错误, 签到器将停止.")
                show_exception(e, regular=False)
                await self.fail()
                message.continue_propagation()
        else:
            message.continue_propagation()

    async def message_handler(self, client: Client, message: Message, type=None):
        """消息处理入口函数."""
        text = message.text or message.caption
        if text:
            for p, k in self._waiting.items():
                if re.search(p, text):
                    k.set()
                    self._waiting[p] = message
        type = type or self.message_type(message)
        if MessageType.TEXT in type:
            await self.on_text(message, message.text)
        if MessageType.CAPTION in type:
            await self.on_text(message, message.caption)
        if MessageType.CAPTCHA in type:
            await self.on_photo(message)

    def message_type(self, message: Message):
        """分析传入消息的类型为验证码或文字."""
        if message.photo:
            if message.caption:
                if self.bot_checkin_caption_pat:
                    if re.search(self.bot_checkin_caption_pat, message.caption):
                        return MessageType.CAPTCHA
                    else:
                        return MessageType.CAPTION
                else:
                    return MessageType.CAPTCHA
            else:
                return MessageType.CAPTCHA
        elif message.text:
            return MessageType.TEXT

    async def on_photo(self, message: Message):
        """分析分析传入的验证码图片并返回验证码."""
        data = await self.client.download_media(message, in_memory=True)
        image = Image.open(data)
        ocr = await self.get_ocr(self.ocr)
        captcha = (
            ocr.classification(image).translate(str.maketrans("", "", string.punctuation)).replace(" ", "")
        )
        if captcha:
            self.log.debug(f"[gray50]接收验证码: {captcha}.[/]")
            if self.bot_captcha_len and len(captcha) not in to_iterable(self.bot_captcha_len):
                self.log.info(f"签到失败: 验证码低于设定长度, 正在重试.")
                await self.retry()
            else:
                await asyncio.sleep(random.uniform(2, 4))
                await self.on_captcha(message, captcha)
        else:
            self.log.info(f"签到失败: 接收到空验证码, 正在重试.")
            await self.retry()

    async def on_captcha(self, message: Message, captcha: str):
        """
        可修改的回调函数.
            message: 包含验证码图片的消息
            captcha: OCR 识别的验证码
        """
        await message.reply(captcha)

    async def on_text(self, message: Message, text: str):
        """接收非验证码消息时, 检测关键词并确认签到成功或失败, 发送用户提示."""
        if any(s in text for s in to_iterable(self.bot_text_ignore)):
            pass
        elif any(s in text for s in self.bot_account_fail_keywords or default_keywords["account_fail"]):
            self.log.warning(f"签到失败: 账户错误.")
            await self.fail()
        elif any(
            s in text
            for s in self.bot_too_many_tries_fail_keywords or default_keywords["too_many_tries_fail"]
        ):
            self.log.warning(f"签到失败: 尝试次数过多.")
            await self.fail()
        elif any(s in text for s in self.bot_checked_keywords or default_keywords["checked"]):
            self.log.info(f"今日已经签到过了.")
            self.finished.set()
        elif any(s in text for s in self.bot_fail_keywords or default_keywords["fail"]):
            self.log.info(f"签到失败: 验证码错误, 正在重试.")
            await self.retry()
        elif any(s in text for s in self.bot_success_keywords or default_keywords["success"]):
            matches = re.search(self.bot_success_pat, text)
            if matches:
                try:
                    self.log.info(f"[yellow]签到成功[/]: + {matches.group(1)} 分 -> {matches.group(2)} 分.")
                except IndexError:
                    self.log.info(f"[yellow]签到成功[/]: 当前/增加 {matches.group(1)} 分.")
            else:
                matches = re.search(r"\d+", text)
                if matches:
                    self.log.info(f"[yellow]签到成功[/]: 当前/增加 {matches.group(0)} 分.")
                else:
                    self.log.info(f"[yellow]签到成功[/].")
            self.finished.set()
        else:
            self.log.warning(f"接收到异常返回信息: {text}")

    async def retry(self):
        """执行重试, 重新发送签到指令."""
        self._retries += 1
        if self._retries <= self.retries:
            await asyncio.sleep(self.bot_retry_wait)
            await self.send_checkin(retry=True)
        else:
            self.log.warning("超过最大重试次数.")
            self.finished.set()

    async def fail(self):
        """设定签到器失败."""
        self.finished.set()
        self._retries = float("inf")

    async def wait_until(self, pattern: str = ".", timeout: float = None):
        """等待特定消息出现."""
        self._waiting[pattern] = e = asyncio.Event()
        try:
            await asyncio.wait_for(e.wait(), timeout)
        except asyncio.TimeoutError:
            return None
        else:
            msg: Message = self._waiting[pattern]
            return msg


class AnswerBotCheckin(BotCheckin):
    """签到类, 用于按钮模式签到."""

    bot_checkin_button_pat: str = None  # 所有按键需要满足的 regex 条件

    def __init__(self, *args, **kw):
        super().__init__(*args, **kw)
        self.mutex = asyncio.Lock()  # 实例变量异步锁
        self.operable = asyncio.Condition(self.mutex)  # 当 message 被设定时提醒等待的异步线程.
        self.message: Message = None  # 存储可用于回复的带按钮的消息.

    async def walk_history(self, limit=0):
        """处理 limit 条历史消息, 并检测是否有验证码或按钮."""
        try:
            answer = None
            captcha = None
            async for m in self.client.get_chat_history(self.chat_name or self.bot_username, limit=limit):
                if MessageType.ANSWER in self.message_type(m):
                    answer = answer or m
                if MessageType.CAPTCHA in self.message_type(m):
                    captcha = captcha or m
                if answer and captcha:
                    break
            else:
                return False
            await self.on_answer(answer)
            await self.on_photo(captcha)
            return True
        except Exception as e:
            self.log.warning("读取历史消息失败, 将不再读取历史消息.")
            show_exception(e)
            return False

    def get_keys(self, message: Message):
        """获得所有按钮信息."""
        reply_markup = message.reply_markup
        if isinstance(reply_markup, InlineKeyboardMarkup):
            return [k.text for r in reply_markup.inline_keyboard for k in r]
        elif isinstance(reply_markup, ReplyKeyboardMarkup):
            return [k.text for r in reply_markup.keyboard for k in r]

    def is_valid_answer(self, message: Message):
        """确认消息是回复按钮消息."""
        if not message.reply_markup:
            return False
        if self.bot_checkin_button_pat:
            for k in self.get_keys(message):
                if not re.search(self.bot_checkin_button_pat, k):
                    return False
        return True

    def message_type(self, message: Message):
        """分析传入消息的类型为验证码或文字."""
        if self.is_valid_answer(message):
            return MessageType.ANSWER | super().message_type(message)
        else:
            return super().message_type(message)

    async def message_handler(self, client: Client, message: Message, type=None):
        """消息处理入口函数."""
        type = type or self.message_type(message)
        if MessageType.ANSWER in type:
            await self.on_answer(message)
        await super().message_handler(client, message, type=type)

    async def on_answer(self, message: Message):
        """当检测到带按钮的用于回答的消息时保存其信息."""
        async with self.mutex:
            if self.message:
                if self.message.date > message.date:
                    return
                else:
                    self.message = message
            else:
                self.message = message
                self.operable.notify()

    async def on_captcha(self, message: Message, captcha: str):
        """
        可修改的回调函数.
        参数:
            message: 包含验证码图片的消息
            captcha: OCR 识别的验证码
        """
        async with self.operable:
            if not self.message:
                await self.operable.wait()
            match = [(k, fuzz.ratio(k, captcha)) for k in self.get_keys(self.message)]
            max_k, max_r = max(match, key=lambda x: x[1])
            if max_r < 75:
                self.log.info(f'未能找到对应 "{captcha}" 的按键, 正在重试.')
                await self.retry()
            else:
                await self.message.click(max_k)
